# -*- coding: utf-8 -*-
"""
Created on Sun Jul 06 19:09:58 2014

@author: Morten
"""
import os
from pathlib import Path

import numpy as np
import pytest

from anypytools.tools import (AnyPyProcessOutput, AnyPyProcessOutputList,
                              _parse_data, array2anyscript, define2str,
                              get_anybodycon_path, parse_anybodycon_output,
                              path2str)


@pytest.yield_fixture(scope="module")
def fixture():
    yield True


def test_define2str():
    assert define2str("test", 2) == '-def test="2"'
    assert define2str("Test", "Main.MyStudy") == '-def Test="Main.MyStudy"'
    assert (
        define2str("test", '"This is a string"')
        == '-def test=---"\\"This is a string\\""'
    )


def test_path2str():
    assert path2str("test", r"C:\hallo.txt") == '-p test=---"C:\\\\hallo.txt"'
    assert path2str("Test", "C:/hallo.txt") == '-p Test=---"C:/hallo.txt"'


def test_array2anyscript():

    mat33 = array2anyscript(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]))
    assert mat33 == "{{1,0,0},{0,1,0},{0,0,1}}"

    mat31 = array2anyscript(np.array([[1, 0, 0]]))
    assert mat31 == "{{1,0,0}}"

    mat13 = array2anyscript(np.array([[1], [0], [0]]))
    assert mat13 == "{{1},{0},{0}}"

    mat3 = array2anyscript(np.array([0.333333333, -1.9999999999, 0.0]))
    assert mat3 == "{0.333333333,-1.9999999999,0}"

    str2 = array2anyscript(np.array(["hallo", "world"]))
    assert str2 == '{"hallo","world"}'


def test_AnyPyProcessOutput():
    out = AnyPyProcessOutputList(
        [
            AnyPyProcessOutput({"AAAA": 1}),
            AnyPyProcessOutput({"AAAA": 2}),
            AnyPyProcessOutput({"AAAA": 3}),
            AnyPyProcessOutput({"AAAA": 4}),
            AnyPyProcessOutput({"AAAA": 5}),
            AnyPyProcessOutput({"AAAA": 6, "ERROR": 0}),
        ]
    )

    assert len(out) == 6
    assert isinstance(out[0], AnyPyProcessOutput)

    # Test slice get
    assert len(out[1:3]) == 2
    assert isinstance(out[0:2][0], AnyPyProcessOutput)

    # Test slice set
    out[:] = [e for e in out if "ERROR" not in e]
    assert len(out) == 5

    assert isinstance(out["A"], np.ndarray)
    assert out["A"].shape == (5,)


def test_AnyPyProcessOutput_to_dataframe():
    data = {
        "Output.Abscissa.t": np.linspace(0, 1, 6),
        "int": 8,
        "float": 0.38,
        "str": "Hello world",
        "one_dim_data": np.ones(6),
        "three_dim_data": np.ones((6, 3)),
        "StringArray": np.array("Hello world"),
        "speciel_length": np.arange(5),
    }
    anypydata = AnyPyProcessOutput(data)
    df = anypydata.to_dataframe()
    assert df.shape == (6, 14)
    df2 = anypydata.to_dataframe(index_var="speciel_length")
    assert df2.shape == (5, 35)
    df3 = anypydata.to_dataframe(index_var=None)
    assert df3.shape == (1, 39)


def test_AnyPyProcessOutput_to_dataframe_interp():
    """Test that we can interpolate the data using 'Output.percent_var'
    from 0 to 100
    """
    time_var = np.linspace(0, 4, 200)
    data = {
        "Output.Abscissa.t": time_var,
        "Output.percent_var": np.linspace(-10, 140, len(time_var)),
        "Output.SomeData": np.sin(time_var),
        "contant": 0.38,
        "contant_str": "Hello world",
    }
    anypydata = AnyPyProcessOutput(data)
    interpolation_values = np.linspace(0, 99, 100)
    df = anypydata.to_dataframe(
        interp_var="Output.percent_var",
        interp_val=interpolation_values,
    )
    assert df.shape == (100, 5)
    assert all(df["Output.percent_var"] == interpolation_values)


def test_AnyPyProcessOutputList_to_dataframe():
    time_len = 6
    no_simulations = 10
    data = {
        "Output.Abscissa.t": np.linspace(0, 1, time_len),
        "three_dim_data": np.ones((time_len, 3)),
    }
    appl = AnyPyProcessOutputList(
        [AnyPyProcessOutput(data) for i in range(no_simulations)]
    )
    df = appl.to_dataframe()
    assert df.shape == (no_simulations * time_len, 4)


def test_get_anybodycon_path():
    abc = get_anybodycon_path()

    assert os.path.exists(abc)


@pytest.mark.parametrize(
    "str_val, expected, expected_dtype",
    [
        (
            """{{{0.9959166, nan, -nan}, {-0.08322453, 0.952727, -0.292207}, {0.0349831, 0.2958367, 0.9545977}},{{0.996183, -0.0689204, 0.05356734}, {0.07839193, 0.9763008, -0.2017211}, {-0.03839513, 0.2051504, 0.9779771}}}""",
            np.array([[[0.9959166, float("nan"), float("nan") ], [-0.08322453, 0.952727, -0.292207], [0.0349831, 0.2958367, 0.9545977]],[[0.996183, -0.0689204, 0.05356734], [0.07839193, 0.9763008, -0.2017211], [-0.03839513, 0.2051504, 0.9779771]]]),
            np.float64,
        ),
        (
            """{{inf, 4.0}, {3.0, -inf}}""",
            np.array([[float("inf"), 4.0],[3.0, float("-inf")]]),
            np.float64,
        ),
        (
            """{1.0, inf, 3.0}""",
            np.array([1.0, float("inf"), 3.0]),
            np.float64,
        ),
    ]
)
def test_parse_anybodydata_arrays(str_val, expected, expected_dtype):
    data_np = _parse_data(str_val)
    assert data_np.dtype == expected_dtype
    assert np.isclose(data_np, expected, equal_nan=True).all()


@pytest.mark.parametrize(
    "str_val, expected, expected_type",
    [
        ("1.0", 1.0, float),
        ('"some_str"', "some_str", str),
        ("1", 1, int),
    ]
  )
def test_parse_anybodydata_scalars(str_val, expected, expected_type):
    out = _parse_data(str_val)
    assert isinstance(out, expected_type)
    assert out == expected


def test_parse_anybodydata_renamed():
    testfile = Path(__file__).parent / "data" / "anybodycon_output.txt"
    raw = testfile.read_text()
    
    data = parse_anybodycon_output(raw)
    assert "Main.SomeOldName" not in data
    assert "Renamed with spaces" in data

    assert "Main.SubFolder.SubOutput" in data
    assert "ReNamedSubFolder.SubOutput" in data


def test_parse_anybodycon_output():
    testfile = Path(__file__).parent / "data" / "anybodycon_output.txt"
    raw = testfile.read_text()

    data = parse_anybodycon_output(raw)

    assert isinstance(data, dict)
    assert "Main.MyOutput" in data
    assert "Main.MyOutput2" in data
    assert "Main.SomeMatrix" in data
    assert "pi" not in data
    assert "Global.pi" in data

    assert isinstance(data["Main.MyOutput"], float)
    assert np.isclose(data["Main.MyOutput"], 42.0)

    assert isinstance(data["Main.MyOutput2"], np.ndarray)
    assert data["Main.MyOutput2"].shape == (3,)

    assert isinstance(data["Global.pi"], float)
    assert data["Global.pi"] == 3.1415926535897931

    assert isinstance(data["Main.SomeMatrix"], np.ndarray)
    assert data["Main.SomeMatrix"].shape == (3, 3)


if __name__ == "__main__":
    os.chdir(Path(__file__).parent)
    pytest.main([str("test_tools.py::test_AnyPyProcessOutputList_to_dataframe")])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# AnyPyTools documentation build configuration file, created by
# sphinx-quickstart on Fri Jul 21 15:33:03 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

from anypytools import __version__ as ANYPYTOOLS_VERSION

import cloud_sptheme


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.githubpages",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "nbsphinx",
    "sphinx.ext.napoleon",
    "IPython.sphinxext.ipython_console_highlighting",
    "cloud_sptheme.ext.autodoc_sections",
    "cloud_sptheme.ext.index_styling",
    "cloud_sptheme.ext.relbar_toc",
    "cloud_sptheme.ext.escaped_samp_literals",
    "cloud_sptheme.ext.issue_tracker",
    "cloud_sptheme.ext.table_styling",
    "myst_parser",
]


napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True

myst_enable_extensions = [
    "colon_fence",
    "deflist",
    "dollarmath",
    "amsmath",
    "html_image",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]


# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "AnyPyTools"
copyright = "2021, Morten Enemark Lund"
author = "Morten Enemark Lund"

github_doc_root = (
    "https://github.com/anybody-research-group/anypytools/tree/master/docs/"
)


# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

if tags.has("DEV_BUILD"):
    release = ANYPYTOOLS_VERSION + ".dev"

# The short X.Y version.
version = ANYPYTOOLS_VERSION.rsplit(".", 1)[0]
# The full version, including alpha/beta/rc tags.
release = ANYPYTOOLS_VERSION


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "**.ipynb_checkpoints", "slides", "Tutorial/00_*.ipynb", "README.md", "paper/paper.md"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------


# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "cloud"

html_theme_path = [cloud_sptheme.get_theme_dir()]


# html_theme = 'sphinxdoc'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "roottarget": "index",
    "max_width": "1250px",
    "minimal_width": "700px",
    "borderless_decor": True,
    "sidebarwidth": "3in",
}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/anypytools_logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/anypytools.ico"

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "AnyPyToolsdoc"


html_logo = "_static/anypytools_logo.png"


nbsphinx_execute = "never"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "AnyPyTools.tex",
        "AnyPyTools Documentation",
        "Morten Enemark Lund",
        "manual",
    )
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "anypytools", "AnyPyTools Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "AnyPyTools",
        "AnyPyTools Documentation",
        author,
        "AnyPyTools",
        "One line description of project.",
        "Miscellaneous",
    )
]


# Autodocumentation Flags
autodoc_member_order = "groupwise"
autoclass_content = "both"
autosummary_generate = []

# Prevent numpy from making silly tables
numpydoc_show_class_members = False



# -*- coding: utf-8 -*-
"""
Created on Sun Jul 06 19:09:58 2014

@author: Morten
"""
import numpy as np
import pytest
from scipy.stats.distributions import norm

import anypytools.macro_commands as mc
from anypytools import AnyMacro

# @pytest.yield_fixture()
# def fixture():
#    mg = MacroGenerator()
#    yield mg


def test_load():
    cmd = mc.Load("main.any", defs={"TEST_STR": '"Test\\string"'})
    assert cmd.get_macro(0) == 'load "main.any" -def TEST_STR=---"\\"Test\\\\string\\""'

    cmd = mc.Load("main.any", defs={"TEST_VAR": "Main.MyStudy"})
    assert cmd.get_macro(0) == 'load "main.any" -def TEST_VAR="Main.MyStudy"'

    cmd = mc.Load("main.any", paths={"TEST_PATH": "c:\\path\\to\\something"})
    assert (
        cmd.get_macro(0)
        == 'load "main.any" -p TEST_PATH=---"c:\\\\path\\\\to\\\\something"'
    )


def test_setvalue():
    c = mc.SetValue("val", 23.1)
    assert c.get_macro(0) == 'classoperation val "Set Value" --value="23.1"'

    c = mc.SetValue("val", -0.123010929395)
    assert c.get_macro(0) == 'classoperation val "Set Value" --value="-0.123010929395"'

    c = mc.SetValue("val", "hallo world")
    assert c.get_macro(0) == 'classoperation val "Set Value" --value="hallo world"'

    c = mc.SetValue("val", [3.0, 4, 5.1])
    assert c.get_macro(0) == 'classoperation val "Set Value" --value="3"'
    assert c.get_macro(1) == 'classoperation val "Set Value" --value="4"'
    assert c.get_macro(2) == 'classoperation val "Set Value" --value="5.1"'
    assert c.get_macro(3) == 'classoperation val "Set Value" --value="3"'

    c = mc.SetValue("val", np.array([1, 2, 3, 4]))
    assert c.get_macro(0) == 'classoperation val "Set Value" --value="{1,2,3,4}"'

    c = mc.SetValue("val", np.array([[1, 0], [0, 1]]))
    assert c.get_macro(0) == 'classoperation val "Set Value" --value="{{1,0},{0,1}}"'


def test_setvalue_random():
    np.random.seed(1)
    fdist = norm(2, [1, 1, 1])

    c = mc.SetValue_random("val", fdist)
    assert c.get_macro(0) == 'classoperation val "Set Value" --value="{2,2,2}"'

    mg = AnyMacro(c)
    macrolist = mg.create_macros_MonteCarlo(2)
    assert macrolist[0][0] == 'classoperation val "Set Value" --value="{2,2,2}"'
    assert (
        macrolist[1][0]
        == 'classoperation val "Set Value" --value="{1.79048215908,2.58380574427,-1.68494766954}"'
    )


def test_dump():
    c = mc.Dump("Main.Study.myvar1")
    assert c.get_macro(0) == 'classoperation Main.Study.myvar1 "Dump"'


def test_export():
    c = mc.Export("Main.Study.myvar1")
    assert c.get_macro(0) == 'classoperation Main.Study.myvar1 "Dump"'

def test_export_rename():
    c = mc.Export("Main.Study.myvar1", "My Var")
    assert c.get_macro(0) == 'print "#### ANYPYTOOLS RENAME OUTPUT: My Var"\nclassoperation Main.Study.myvar1 "Dump"'


def test_savedesign():
    c = mc.SaveDesign("Main.MyStudy.Kinematics", "c:/design.txt")
    assert (
        c.get_macro(0)
        == 'classoperation Main.MyStudy.Kinematics "Save design" --file="c:/design.txt"'
    )


def test_loaddesign():
    c = mc.LoadDesign("Main.MyStudy.Kinematics", "c:/design.txt")
    assert (
        c.get_macro(0)
        == 'classoperation Main.MyStudy.Kinematics "Load design" --file="c:/design.txt"'
    )


def test_savevalues():
    c = mc.SaveValues("c:/design.anyset")
    assert (
        c.get_macro(0) == 'classoperation Main "Save Values" --file="c:/design.anyset"'
    )


def test_savedata():
    c = mc.SaveData("Main.MyStudy", "output.anydata.h5")
    assert (
        c.get_macro(0)
        == 'classoperation Main.MyStudy.Output "Save data" --type="Deep" --file="output.anydata.h5"'
    )


def test_loaddata():
    c = mc.LoadData("Main.MyStudy", "output.anydata.h5")
    assert (
        c.get_macro(0)
        == 'classoperation Main.MyStudy.Output "Load data" --file="output.anydata.h5"'
    )


def test_loadvalues():
    c = mc.LoadValues("c:/design.anyset")
    assert (
        c.get_macro(0) == 'classoperation Main "Load Values" --file="c:/design.anyset"'
    )


def test_updatevalues():
    c = mc.UpdateValues()
    assert c.get_macro(0) == 'classoperation Main "Update Values"'


def test_runoperation():
    c = mc.RunOperation("Main.MyStudy.Kinematics")
    assert c.get_macro(0) == "operation Main.MyStudy.Kinematics\nrun"


def test_macrocommand():
    c = mc.MacroCommand("My macro cmd")
    assert c.get_macro(0) == "My macro cmd"

    c = mc.MacroCommand(["c1", "c2"])
    assert c.get_macro(0) == "c1\nc2"


def test_macros():
    mcr = AnyMacro(number_of_macros=10)
    mcr.append(mc.Load("main.any"))

    macros = mcr.create_macros()

    assert macros[0][0] == 'load "main.any"'
    assert macros[1][0] == 'load "main.any"'
    assert len(macros) == 10


# def test_macro2():
# mcr = AnyMacro([
# mc.Load('main.any'),
# mc.RunOperation('Main.MyStudy.Kinematics')
# ])

# assert str(mcr) == 'kd'

if __name__ == "__main__":
    import pytest

    pytest.main(
        str("test_macros.py ../anypytools/generate_macros.py --doctest-modules")
    )



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 13:38:15 2015

@author: kpr
"""
import os
import shutil
import pytest
from anypytools import AnyPyProcess

demo_model_path = os.path.join(os.path.dirname(__file__), "Demo.Arm2D.any")


def setup_simple_model(tmpdir):
    shutil.copyfile(demo_model_path, str(tmpdir.join("model.main.any")))


@pytest.yield_fixture()
def init_simple_model(tmpdir):
    setup_simple_model(tmpdir)
    with tmpdir.as_cwd():
        yield tmpdir


@pytest.fixture()
def default_macro():
    macro = [
        [
            'load "model.main.any"',
            "operation Main.ArmModelStudy.InverseDynamics",
            "run",
            'classoperation Main.ArmModelStudy.Output.MaxMuscleActivity "Dump"',
            'classoperation Main.ArmModel.GlobalRef.t "Dump"',
        ]
    ] * 5
    return macro


def test_hdf5save(tmpdir, default_macro):
    setup_simple_model(tmpdir)
    tmpdir.chdir()
    app = AnyPyProcess(keep_logfiles=True)
    app.start_macro(default_macro)

    app.save_to_hdf5("test.hdf5", "LungeTLEM1")

    assert os.path.exists("test.hdf5")


if __name__ == "__main__":
    macro = [
        [
            'load "Demo.Arm2D.any"',
            "operation Main.ArmModelStudy.InverseDynamics",
            "run",
            'classoperation Main.ArmModelStudy.Output.MaxMuscleActivity "Dump"',
            'classoperation Main.ArmModel.GlobalRef.t "Dump"',
        ]
    ] * 5
    app = AnyPyProcess(keep_logfiles=True)
    app.start_macro(macro)

    app.save_to_hdf5("test2.hdf5", "Lunge11dec")



# -*- coding: utf-8 -*-
"""
Created on Wed Sep  3 11:03:41 2014

@author: Morten
"""
from anypytools.datautils import read_anyoutputfile


def test_open_anyoutputfile(request):
    testfile = str(request.fspath.new(basename="any_output_file.txt"))
    data, header, constants = read_anyoutputfile(testfile)

    assert constants["Main.Studies.ElbowExtension.Outputfile.SepSign"] == ","
    assert (
        constants["Main.Studies.ElbowExtension.Outputfile." "NumberFormat.FormatStr"]
        == "%22.15e"
    )

    assert "Main.Studies.ElbowExtension.IterationStep" in header

    assert data.shape == (27, 12)



# -*- coding: utf-8 -*-
"""
Created on Sun Jul 06 19:09:58 2014

@author: Morten
"""
import os
import shutil
import pytest
import pathlib


from anypytools.abcutils import AnyPyProcess
from anypytools.abcutils import AnyPyProcessOutputList, AnyPyProcessOutput

demo_model_path = os.path.join(os.path.dirname(__file__), "Demo.Arm2D.any")


def setup_simple_model(tmpdir):
    shutil.copyfile(demo_model_path, str(tmpdir.join("model.main.any")))


def setup_models_in_subdirs(tmpdir, number_of_models=5):
    for i in range(number_of_models):
        subdir = tmpdir.mkdir("model" + str(i))
        setup_simple_model(subdir)


@pytest.yield_fixture()
def init_simple_model(tmpdir):
    setup_simple_model(tmpdir)
    with tmpdir.as_cwd():
        yield tmpdir


@pytest.yield_fixture()
def default_macro():
    macro = [['load "model.main.any"', "operation Main.ArmModelStudy.InverseDynamics"]]
    yield macro


class TestAnyPyProcess:
    def test_logfile_persistance(self, init_simple_model, default_macro):
        app = AnyPyProcess(silent=True, keep_logfiles=True)
        output = app.start_macro(default_macro)
        assert os.path.isfile(output[0]["task_logfile"])

    def test_explicit_logfiles_str(self, init_simple_model, default_macro):
        logfile = "test.log"
        app = AnyPyProcess(silent=True, keep_logfiles=True)
        output = app.start_macro(default_macro, logfile=logfile)
        assert os.path.isfile(logfile)

    def test_explicit_logfiles_path(self, init_simple_model, default_macro):
        logfile = pathlib.Path("test.log")
        app = AnyPyProcess(silent=True, keep_logfiles=True)
        output = app.start_macro(default_macro, logfile=logfile)
        assert os.path.isfile(logfile)

    def test_with_gui(self, init_simple_model, default_macro):
        app = AnyPyProcess(silent=True, use_gui=False)
        default_macro[0].append("run")
        output = app.start_macro(default_macro)
        assert not "ERROR" in output[0]

    def test_explicit_logfiles_multiple(self, init_simple_model):
        logfile = "test.log"
        macro = [
            ['load "model.main.any"', "operation Main.ArmModelStudy.InverseDynamics"],
            ['load "model.main.any"', "operation Main.ArmModelStudy.InverseDynamics"],
            ['load "model.main.any"', "operation Main.ArmModelStudy.InverseDynamics"],
        ]
        app = AnyPyProcess(silent=True, keep_logfiles=True)
        output = app.start_macro(macro, logfile=logfile)
        assert os.path.isfile("test_0.log")
        assert os.path.isfile("test_1.log")
        assert os.path.isfile("test_2.log")

    def test_macro_with_erros(self, init_simple_model, default_macro):
        app = AnyPyProcess(silent=True, ignore_errors=["Unresolved object"])
        macro = [
            ['load "model.main.any"', "operation Main.ArmModelStudy.InverseDynamics"],
            ['load "model.main.any"', "operation Main.NonExistentOpeation"],
            ['load "not_a_model.any"', "operation Main.ArmModelStudy.InverseDynamics"],
        ]

        output = app.start_macro(macro)

        assert not "ERROR" in output[0]
        assert not "ERROR" in output[1]
        assert "ERROR" in output[2]

    def test_start_macro(self, init_simple_model, default_macro):
        app = AnyPyProcess(silent=True)

        default_macro[0].extend(
            [
                'classoperation Main.ArmModelStudy.Output.MaxMuscleActivity "Dump"',
                'classoperation Main.ArmModel.GlobalRef.t "Dump"',
            ]
        )

        output = app.start_macro(default_macro)

        assert len(output) == 1
        assert "Main.ArmModelStudy.Output.MaxMuscleActivity" in output[0]
        assert "Main.ArmModel.GlobalRef.t" in output[0]
        assert "ERROR" not in output[0]

    def test_start_macro_with_task_info(self, init_simple_model, default_macro):
        app = AnyPyProcess(silent=True)

        default_macro[0].extend(
            [
                'classoperation Main.ArmModelStudy.Output.MaxMuscleActivity "Dump"',
                'classoperation Main.ArmModel.GlobalRef.t "Dump"',
            ]
        )

        output = app.start_macro(default_macro)

        assert len(output) == 1
        assert "task_macro_hash" in output[0]
        assert "task_id" in output[0]
        assert "task_work_dir" in output[0]
        assert "task_name" in output[0]
        assert "task_processtime" in output[0]
        assert "task_macro" in output[0]
        assert "task_logfile" in output[0]

    def test_start_macro_subdirs(self, tmpdir, default_macro):

        number_of_models = 5
        setup_models_in_subdirs(tmpdir, number_of_models)

        app = AnyPyProcess(silent=True)
        with tmpdir.as_cwd():
            output = app.start_macro(
                default_macro, search_subdirs="main.any", logfile="test.log"
            )

        assert len(output) == number_of_models
        for result in output:
            assert "ERROR" not in result

    def test_start_macro_generator_input(self, init_simple_model, default_macro):
        n_macros = 5

        def generate_macros():
            """Generator that returns macros"""
            for i in range(n_macros):
                yield default_macro[0]

        app = AnyPyProcess(silent=True)
        macros_gen = generate_macros()
        output = app.start_macro(macros_gen)

        assert len(output) == n_macros
        for result in output:
            assert "ERROR" not in result

    def test_start_macro_multple_folders_and_macros(self, tmpdir, default_macro):
        number_of_models = 3
        number_of_macros = 3
        setup_models_in_subdirs(tmpdir, number_of_models)
        folderlist = [str(_) for _ in tmpdir.listdir()]
        macrolist = default_macro * number_of_macros

        with tmpdir.as_cwd():
            app = AnyPyProcess()
            output = app.start_macro(macrolist, folderlist)

        assert len(output) == len(folderlist) * len(macrolist)
        for result in output:
            assert "ERROR" not in result

    def test_output_access_ragged(self, init_simple_model):
        macro = [
            [
                'load "model.main.any" -def N_STEP="10"',
                "operation Main.ArmModelStudy.InverseDynamics",
                'classoperation Main.ArmModelStudy.Output.MaxMuscleActivity "Dump"',
                'classoperation Main.ArmModel.GlobalRef.t "Dump"',
            ],
            [
                'load "model.main.any" -def N_STEP="20"',
                "operation Main.ArmModelStudy.InverseDynamics",
                'classoperation Main.ArmModelStudy.Output.MaxMuscleActivity "Dump"',
                'classoperation Main.ArmModel.GlobalRef.t "Dump"',
            ],
        ]

        app = AnyPyProcess()
        output = app.start_macro(macro)
        output["GlobalRef.t"]
        output["MaxMuscleActivity"]

    def test_output_access(self, init_simple_model):
        macro = [
            [
                'load "model.main.any" -def N_STEP="5"',
                "operation Main.ArmModelStudy.InverseDynamics",
                'classoperation Main.ArmModelStudy.Output.MaxMuscleActivity "Dump"',
                'classoperation Main.ArmModel.GlobalRef.t "Dump"',
            ]
        ] * 5

        app = AnyPyProcess()
        output = app.start_macro(macro)
        output["GlobalRef.t"]
        assert (
            output["MaxMuscleActivity"].shape
            == output["Main.ArmModelStudy.Output.MaxMuscleActivity"].shape
        )
        assert isinstance(output[1:3], AnyPyProcessOutputList)

    def test_output_save_load(self, init_simple_model):
        macro = [
            [
                'load "model.main.any" -def N_STEP="10"',
                "operation Main.ArmModelStudy.InverseDynamics",
                'classoperation Main.ArmModelStudy.Output.MaxMuscleActivity "Dump"',
                'classoperation Main.ArmModel.GlobalRef.t "Dump"',
            ]
        ]

        app = AnyPyProcess()
        app.start_macro(macro)
        app.save_results("test.db")

        reloaded = app.load_results("test.db")
        reloaded["Main.ArmModel.GlobalRef.t"]

    def test_restart_macro(self, init_simple_model):
        macro = [
            [
                'load "model.main.any"',
                'classoperation Main.ArmModelStudy.Output.MaxMuscleActivity "Dump"',
                'classoperation Main.ArmModel.GlobalRef.t "Dump"',
            ],
            [
                'load "model.main.any"',
                'classoperation Main.ArmModelStudy.Output.MaxMuscleActivity "Dump"',
                'classoperation Main.ArmModel.GlobalRef.t "Dump"',
            ],
        ]

        app = AnyPyProcess()

        output = app.start_macro(macro)
        output = app.start_macro()

        output[0]["ERROR"] = ["SOME ERROR"]

        output = app.start_macro(output)

        for result in output:
            assert "ERROR" not in result

    def test_output_dataframe_with_time(self, init_simple_model):

        macro = [
            [
                'load "model.main.any"',
                "operation Main.ArmModelStudy.InverseDynamics",
                "run",
                'classoperation Main.ArmModelStudy.MaxMuscleActivity "Dump"',
                'classoperation Main.ArmModelStudy.Output.Model.Segs.LowerArm.ElbowNode.r "Dump"',
                'classoperation Main.ArmModelStudy.Ekin "Dump"',
                'classoperation Main.ArmModelStudy.Output.Abscissa.t "Dump"',
            ],
        ]
        app = AnyPyProcess(silent=True)
        output: AnyPyProcessOutput = app.start_macro(macro)[0]
        df = output.to_dataframe()
        assert df.shape == (100, 12)

    def test_output_dataframe_without_time(self, init_simple_model):

        macro = [
            [
                'load "model.main.any"',
                "operation Main.ArmModelStudy.InverseDynamics",
                "run",
                'classoperation Main.ArmModelStudy.MaxMuscleActivity "Dump"',
                'classoperation Main.ArmModelStudy.Output.Model.Segs.LowerArm.ElbowNode.r "Dump"',
                'classoperation Main.ArmModelStudy.Ekin "Dump"',
                'classoperation Main.Epot "Dump"',
            ],
        ]
        app = AnyPyProcess(silent=True)

        output: AnyPyProcessOutput = app.start_macro(macro)[0]
        df = output.to_dataframe(index_var=None)
        assert df.shape == (1, 308)
        assert df.index.name == None


if __name__ == "__main__":
    os.chdir(pathlib.Path(__file__).parent)
    pytest.main(
        [str("test_abcutils.py::TestAnyPyProcess::test_output_dataframe_with_time")]
    )



# -*- coding: utf-8 -*-
"""

"""
import os
from anypytools import h5py_wrapper

h5test_file = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "test.anydata.h5"
)


def test_file():
    with h5py_wrapper.File(h5test_file) as f:
        t1 = f["/Output/MomentArm"]  # Standard h5py notation
        t2 = f["Output.MomentArm"]  # dot notation
        t3 = f["Main.MyStudy.Output.Model.Knee.Pos"]  # dot notation with full path


def test_group():
    with h5py_wrapper.File(h5test_file) as f:
        h5group = f["/Output"]
        t1 = h5group["MomentArm"]  # Standard h5py notation
        t2 = h5group["Model/Knee/Pos"]
        t3 = h5group["Model.Knee.Pos"]



# coding: utf-8
"""
Adaapted from https://stackoverflow.com/a/56632466

This module provides a JobPopen class that is a subclass of the Popen class from the subprocess module.

"""

import subprocess
from subprocess import Popen
from threading import RLock

import win32api
import win32job
import win32process
import pywintypes


class JobPopen(Popen):
    """Start a process in a new Win32 job object.

    This `subprocess.Popen` subclass takes the same arguments as Popen and
    behaves the same way. In addition to that, created processes will be
    assigned to a new anonymous Win32 job object on startup, which will
    guarantee that the processes will be terminated by the OS as soon as
    either the Popen object, job object handle or parent Python process are
    closed.
    """

    class _winapijobhandler(object):
        """Patches the native CreateProcess function in the subprocess module
        to assign created threads to the given job"""

        def __init__(self, oldapi, job):
            self._oldapi = oldapi
            self._job = job

        def __getattr__(self, key):
            if key != "CreateProcess":
                return getattr(self._oldapi, key)  # Any other function is run as before
            else:
                return self.CreateProcess  # CreateProcess will call the function below

        def CreateProcess(self, *args, **kwargs):
            hp, ht, pid, tid = self._oldapi.CreateProcess(*args, **kwargs)
            try:
                win32job.AssignProcessToJobObject(self._job, hp)
            except BaseException as e:  # to catch pywintypes.error
                if e.args == (6, "AssignProcessToJobObject", "The handle is invalid."):
                    # Try to ignore an error can occur randomly sometimes.
                    pass
                else:
                    raise e
            win32process.ResumeThread(ht)
            return hp, ht, pid, tid

    def __init__(self, *args, **kwargs):
        """Start a new process using an anonymous job object. Takes the same arguments as Popen"""

        # Create a new job object
        self._win32_job = self._create_job_object()

        # Temporarily patch the subprocess creation logic to assign created
        # processes to the new job, then resume execution normally.
        CREATE_SUSPENDED = 0x00000004
        kwargs.setdefault("creationflags", 0)
        kwargs["creationflags"] |= CREATE_SUSPENDED
        with RLock():
            _winapi = subprocess._winapi  # Python 3
            _winapi_key = "_winapi"
            try:
                setattr(
                    subprocess,
                    _winapi_key,
                    JobPopen._winapijobhandler(_winapi, self._win32_job),
                )
                super(JobPopen, self).__init__(*args, **kwargs)
            finally:
                setattr(subprocess, _winapi_key, _winapi)

    def _create_job_object(self):
        """Create a new anonymous job object"""
        hjob = win32job.CreateJobObject(None, "")
        extended_info = win32job.QueryInformationJobObject(
            hjob, win32job.JobObjectExtendedLimitInformation
        )
        extended_info["BasicLimitInformation"][
            "LimitFlags"
        ] = win32job.JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE
        win32job.SetInformationJobObject(
            hjob, win32job.JobObjectExtendedLimitInformation, extended_info
        )
        return hjob

    def _close_job_object(self, hjob):
        """Close the handle to a job object, terminating all processes inside it"""
        if self._win32_job:
            win32api.CloseHandle(self._win32_job)
            self._win32_job = None

    # This ensures that no remaining subprocesses are found when the process
    # exits from a `with JobPopen(...)` block.
    def __exit__(self, exc_type, value, traceback):
        super(JobPopen, self).__exit__(exc_type, value, traceback)
        self._close_job_object(self._win32_job)

    # Python does not keep a reference outside of the parent class when the
    # interpreter exits, which is why we keep it here.
    _Popen = subprocess.Popen

    def __del__(self):
        self._Popen.__del__(self)
        self._close_job_object(self._win32_job)



# -*- coding: utf-8 -*-
"""AnyPyTools library."""
import logging
import os
import platform
import sys

if "FOR_DISABLE_CONSOLE_CTRL_HANDLER" not in os.environ:
    os.environ["FOR_DISABLE_CONSOLE_CTRL_HANDLER"] = "1"

from anypytools import macro_commands
from anypytools.abcutils import AnyPyProcess, execute_anybodycon
from anypytools.macroutils import AnyMacro
from anypytools.tools import (
    ABOVE_NORMAL_PRIORITY_CLASS,
    BELOW_NORMAL_PRIORITY_CLASS,
    IDLE_PRIORITY_CLASS,
    NORMAL_PRIORITY_CLASS,
)

logger = logging.getLogger("abt.anypytools")
logger.addHandler(logging.NullHandler())


__all__ = [
    "datautils",
    "h5py_wrapper",
    "AnyPyProcess",
    "AnyMacro",
    "macro_commands",
    "print_versions",
    "execute_anybodycon",
    "ABOVE_NORMAL_PRIORITY_CLASS",
    "BELOW_NORMAL_PRIORITY_CLASS",
    "IDLE_PRIORITY_CLASS",
    "NORMAL_PRIORITY_CLASS",
]

__version__ = "1.18.0"


def print_versions():
    """Print all the versions of software that AnyPyTools relies on."""
    import numpy as np
    import scipy as sp

    print("-=" * 38)
    print(f"AnyPyTools version: {__version__}")
    print(f"NumPy version: {np.__version__}")
    print(f"SciPy version: {sp.__version__}")
    print(f"Python version: {sys.version}")
    (sysname, _, release, version, machine, processor) = platform.uname()
    print(f"Platform: {sysname}-{release}-{machine} ({version})")
    if not processor:
        processor = "not recognized"
    print(f"Processor: {processor}")
    print(f"Byte-ordering: {sys.byteorder}")
    print("-=" * 38)



from anypytools import AnyPyProcess
from anypytools.macro_commands import Load, RunOperation

app = AnyPyProcess(num_processes=3)

macro = [Load("main.any"), RunOperation("Main.Study.InverseDynamics")]

app.start_macro(macro, search_subdirs="model[1-9].*main.any")
